/* Copyright 2017, 2018, 2022 Kevin Ryde

   This file is part of nautyextra.

   nautyextra is free software; you can redistribute it and/or modify it
   under the terms of the GNU General Public License as published by the
   Free Software Foundation; either version 3, or (at your option) any later
   version.

   nautyextra is distributed in the hope that it will be useful, but WITHOUT
   ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
   FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
   more details.

   You should have received a copy of the GNU General Public License along
   with nautyextra.  See the file COPYING.  If not, see
   <http://www.gnu.org/licenses/>.
*/

#include <nauty/nauty.h>
#include <nauty/nausparse.h>

#ifndef  _NAUTYEXTRA_WIP_H_
#define  _NAUTYEXTRA_WIP_H_

#ifdef __cplusplus
extern "C" {
#endif

#ifndef NAUTYEXTRA_TYPE_ULL
#define NAUTYEXTRA_TYPE_ULL   unsigned long long
#endif

#ifndef DELONEEDGE

#define DELONEARC1(g,v,w,m) DELELEMENT1(GRAPHROW1(g,v,m),w)
#define DELONEEDGE1(g,v,w,m) { DELONEARC1(g,v,w,m); DELONEARC1(g,w,v,m); }

#define DELONEARC0(g,v,w,m) DELELEMENT0(GRAPHROW0(g,v,m),w)
#define DELONEEDGE0(g,v,w,m) { DELONEARC0(g,v,w,m); DELONEARC0(g,w,v,m); }

#if  (MAXM==1) && defined(ONE_WORD_SETS)
#define DELONEARC DELONEARC1
#define DELONEEDGE DELONEEDGE1
#else
#define DELONEARC DELONEARC0
#define DELONEEDGE DELONEEDGE0
#endif

#endif

#if 0
#define NAUTYEXTRA_DELETE_ARC(g,u,v,m)  DELONEARC(g,u,v,m)
#define NAUTYEXTRA_DELETE_EDGE(g,u,v,m)  DELONEEDGE(g,u,v,m)
#endif

#if 0
#define NAUTYEXTRA_DELETE_ARC(g,u,v,m)  DELELEMENT(GRAPHROW(g2,u,m),v)
#define NAUTYEXTRA_DELETE_EDGE(g,u,v,m)                       \
  do {                                  \
    NAUTYEXTRA_DELETE_ARC(g,u,v,m);     \
    NAUTYEXTRA_DELETE_ARC(g,v,u,m);     \
  } while (0)
#endif

extern int nautyextra_want_progress;
extern int nautyextra_progress_reps;
int nautyextra_progress(void);
#define NAUTYEXTRA_PROGRESS \
  (nautyextra_want_progress && nautyextra_progress())

extern int nautyextra_view_synchronous;
void nautyextra_view (const graph *g, int m, int n);
void nautyextra_view_sg (const sparsegraph *sg);
void nautyextra_view_vpar (const int *vpar, int n);
void nautyextra_view_g6 (const char *str);
int nautyextra_hog_grep(const graph *g, int m, int n);

void nautyextra_print_matrix (const graph *g, int m,int n);
void nautyextra_print_vpar_gp (const int *vpar, int n);

void nautyextra_sg_realloc_nv (sparsegraph *sg, int nv);
void nautyextra_sg_realloc_elen (sparsegraph *sg, size_t nde);
void nautyextra_disconnect_vertex (graph *g, int m, int n, int v);

int nautyextra_set_nthelement1 (setword ss, int n);
int nautyextra_set_nthelement (const set *ss, int n);
void nautyextra_set_union(set *dst, const set *src, int m);
int nautyextra_set_is_subset (const set *subset, const set* superset, int m);

int nautyextra_is_well_covered_sg (const sparsegraph *sg);

int nautyextra_max_stabilizing_sequence (const graph *g, int m, int n, int *seq);
int nautyextra_is_semistable_at_vertex (const graph *g, int m, int n, int v);
int nautyextra_is_completely_semistable (const graph *g, int m, int n);
int nautyextra_is_stable (const graph *g, int m, int n);


struct nautyextra_disjoint_domnum_and_count {
  int disjoint_domnum;
  NAUTYEXTRA_TYPE_ULL count;
};
int nautyextra_disjoint_domnum_vpar (const int *vpar, int n);
void nautyextra_disjoint_domnum_and_count_vpar (const int *vpar, int n, struct nautyextra_disjoint_domnum_and_count *info);

int nautyextra_is_invdomset_sg (const sparsegraph *sg, const set *ss);
NAUTYEXTRA_TYPE_ULL nautyextra_invdomsets_count_sg (const sparsegraph *sg);
struct nautyextra_invdomnum_and_count {
  int invdomnum;
  NAUTYEXTRA_TYPE_ULL count;
};
void nautyextra_invdomnum_and_count_sg (const sparsegraph *sg, struct nautyextra_invdomnum_and_count *info);

int nautyextra_is_semitotdomset_sg (const sparsegraph *sg, const set *ss);
struct nautyextra_semitotdomnum_and_count {
  int semitotdomnum;
  NAUTYEXTRA_TYPE_ULL count;
};
void nautyextra_semitotdomnum_and_count_sg (const sparsegraph *sg, struct nautyextra_semitotdomnum_and_count *info);
int nautyextra_semitotdomnum_vpar (const int *vpar, int n);

int nautyextra_is_perfect_domset_sg (const sparsegraph *sg, const set *ss);
struct nautyextra_perfect_domnum_and_count {
  int perfect_domnum;
  NAUTYEXTRA_TYPE_ULL count;
};
void nautyextra_perfect_domnum_and_count_sg (const sparsegraph *sg, struct nautyextra_perfect_domnum_and_count *info);
int nautyextra_perfect_domnum_vpar (const int *vpar, int n);

int nautyextra_loopcount_sg (const sparsegraph *sg);
int nautyextra_leaf_count_vpar (const int *vpar, int n);
int nautyextra_nonleaf_count_vpar (const int *vpar, int n);
int nautyextra_childless_count_vpar (const int *vpar, int n);
int nautyextra_parent_count_vpar (const int *vpar, int n);
int nautyextra_height_num_components_vpar (const int *vpar, int n);
void nautyextra_compnum_array_vpar (const int *vpar, int n, int *components);

int nautyextra_num_edges (const graph *g, int m, int n);
int nautyextra_is_arc_sg (const sparsegraph *sg, int u, int v);
int nautyextra_is_multiedged_sg (const sparsegraph *sg);

void nautyextra_degree_histogram_vpar (const int *vpar, int n, int *hist);

int nautyextra_Szeged_index (const graph *g, int m, int n);

void nautyextra_sg_to_vpar (const sparsegraph *sg, int *vpar, int root);
void nautyextra_vpar_to_graph (const int *vpar, graph *g, int m, int n, int directed);
void nautyextra_vpar_to_sg (const int *vpar, int n, sparsegraph *sg, int directed);
const char *nautyextra_vpar_to_g6 (const int *vpar, int n);
const char *nautyextra_vpar_to_s6 (const int *vpar, int n);
const char *nautyextra_vpar_to_d6 (const int *vpar, int n, int directed);

int nautyextra_any_duplicated_leaf_vpar (const int *vpar, int n);
int nautyextra_limb_height_vpar (const int *vpar, int n);
int nautyextra_limb_length_vpar (const int *vpar, int n);

int nautyextra_is_cyclic (const graph *g, int m, int n);
int nautyextra_is_cyclic_vpar (const int *vpar, int n);
int nautyextra_is_connected_vpar (const int *vpar, int n);
int nautyextra_is_cut_edge (const graph *g, int m, int n, int u, int v);
int nautyextra_cut_edge_count (const graph *g, int m, int n);

void nautyextra_depths_vpar (const int *vpar, int n, int *depths);
int nautyextra_height_vpar (const int *vpar, int n);
int nautyextra_2height_vpar (const int *vpar, int n);
void nautyextra_subtree_heights_vpar (const int *vpar, int n, int *heights);

int nautyextra_eccentricity_of_vertex (const graph *g, int m, int n, int v, int *destination_ptr);
int nautyextra_eccentricity_of_vertex_sg (const sparsegraph *sg, int v, int *destination_ptr);
void nautyextra_eccentricities_vpar (const int *vpar, int n, int *ecc);
void nautyextra_centres_vpar (const int *vpar, int n, int *c1p, int *c2p);
void nautyextra_path_centres_vpar (const int *vpar, int v1, int v2, int p, int *c1p, int *c2p);

struct nautyextra_Wiener_info {
  NAUTYEXTRA_TYPE_ULL W;
  int diameter;
  int diameter_from;
  int diameter_to;
  int num;
};
void nautyextra_Wiener_index (const graph *g, int m, int n, int terminal, struct nautyextra_Wiener_info *out);
void nautyextra_Wiener_index_sg (const sparsegraph *sg, int terminal, struct nautyextra_Wiener_info *out);
void nautyextra_Wiener_index_vpar (const int *vpar, int n, int terminal, struct nautyextra_Wiener_info *out);

int nautyextra_aresame_vpar (const int *vpar1, const int *vpar2, int n);
void nautyextra_canonise(const graph *g_in, int m, int n, graph *g_out);
sparsegraph *nautyextra_canonise_sg(const sparsegraph *sg_in, sparsegraph *sg_out);
char *nautyextra_canonical_str (const graph *g, int m, int n, int format);
int nautyextra_is_isomorphic (const graph *g1, const graph *g2, int m, int n);
int nautyextra_is_isomorphic_sg (const sparsegraph *sg1, const sparsegraph *sg2);
int nautyextra_is_isomorphic_gsg (const graph *g, const sparsegraph *sg, int m, int n);
int nautyextra_is_isomorphic_vpar (const int *vpar1, const int *vpar2, int n, int directed, int *perm);
int nautyextra_is_isomorphic_vpar2 (const int *vpar1, int n1, const int *vpar2, int n2, int directed);
int nautyextra_is_isomorphic_sg_vpar (const sparsegraph *sg, const int *vpar, int n);
int nautyextra_is_selfcomplement (const graph *g, int m, int n);

int nautyextra_is_matching_vpar_set (const int *vpar, int n, const set *ss);

void nautyextra_max_ancestor_array_vpar (const int *vpar, int n, int *max_ancestor);
void nautyextra_seq_upascend (const int *vpar, int n, int ahead, int *seq);
void nautyextra_from_upascend (const int *vec, int n, int ahead, int *vpar);
void nautyextra_seq_subtree_height_vpar (const int *vpar, int n, int *seq);
void nautyextra_seq_bucket_sort (const int *keys, int n, int *seq);

int nautyextra_can_count (int n, int forest);
void nautyextra_can_init(int *can, int n, int forest);
int nautyextra_can_next(int *can, int n, int forest);

void nautyextra_make_path (graph *g, int m, int n, int directed);
void nautyextra_make_star (graph *g, int m, int n, int directed);
void nautyextra_make_star_arms2_vpar (int *vpar, int n);
int nautyextra_make_bistar_n (int a, int b, int length);
void nautyextra_make_bistar_vpar (int *vpar, int a, int b, int length);
void nautyextra_make_complete (graph *g, int m, int n, int directed);
void nautyextra_make_complete_tree_vpar (int *vpar, int c, int h);
int  nautyextra_make_complete_tree_size (int c, int h);
void nautyextra_make_cycle (graph *g, int m, int n, int directed);
void nautyextra_make_cycle_sg (sparsegraph *sg, int n);
void nautyextra_make_grid (graph *g, int m, int n, int xsize, int ysize);
void nautyextra_make_hypercube (graph *g, int m, int n, int d);
void nautyextra_make_complete_bipartite (graph *g, int m, int n, int a, int b);
void nautyextra_make_Petersen (graph *g, int m, int n, int k);
void nautyextra_make_Mobius_ladder (graph *g, int m, int n);
void nautyextra_make_most_indomsets (graph *g, int m, int n);

int nautyextra_join_vpar (int *vpar, int n, int v, int *sub_vpar, int sub_n);
int nautyextra_join_paths_vpar (int *vpar, int n, int s);

void nautyextra_star_replacement(graph *g_in,  int m_in,  int n_in,
                                 graph *g_out, int m_out,  int s);
void nautyextra_star_replacement_sg (sparsegraph *sg_in, sparsegraph *sg_out, int s);

void nautyextra_ranrelabel (const graph *g_in, int m, int n, graph *g_out);
void nautyextra_relabel_swap_vpar_copy (const int *vpar_in, int n, int u, int v, int *vpar_out);

void nautyextra_to_undirected (graph *g, int m, int n);

void nautyextra_random_arc (const graph *g, int m, int n, int *up, int *vp);
void nautyextra_random_arc_sg (const sparsegraph *sg, int *up, int *vp);
int nautyextra_nth_arc (const graph *g, int m, int n, int arc, int *up, int *vp);

int nautyextra_triangle_search (const graph *g, int m, int n, int (*callback)(const graph*,int,int, int,int,int, void *), void *data);
int nautyextra_triangle_is_even (const graph *g, int m, int a, int b, int c);

int nautyextra_diamond_search (const graph *g, int m, int n, int induced, int (*callback)(const graph*,int,int, int,int,int,int, void *), void *data);
int nautyextra_diamond_count (const graph *g, int m, int n, int induced);
int nautyextra_diamond_is_both_odd (const graph *g, int m, int a, int b, int c, int d);

void nautyextra_linegraph (const graph *g_in, int m_in, int n_in, graph *g_out, int m_out, int n_out);
int nautyextra_is_linegraph1 (const graph *g, int n);
int nautyextra_is_linegraph (const graph *g, int m, int n);

int nautyextra_is_subgraph (const graph *sub_g, int sub_m, int sub_n, const graph *g, int m, int n, int induced);

int nautyextra_is_clique1 (const graph *g, int n, setword ss);
int nautyextra_is_clique (const graph *g, int m, int n, const set *ss);
int nautyextra_clique_number (const graph *g, int m, int n);
struct nautyextra_similar_info {
  graph *similar;
  graph *removal_similar;
  graph *pseudosimilar;
  int   pseudosimilar_clique_size;
  int   numorbits;
  int   orbit_max;
};
void nautyextra_similar (graph *g, int m, int n, struct nautyextra_similar_info *info);

int nautyextra_regular_degree (const graph *g, int m, int n);
void nautyextra_distances_matrix (const graph *g, int m, int n, int *distances);
int nautyextra_is_distance_regular (const graph *g, int m, int n, int *intersection_array);
void nautyextra_find_dist_counts(const graph *g, int m, int n, int v, const int *dist, unsigned long long *count);
int nautyextra_total_distance_of_vertex (const graph *g, int m, int n, int v, const set *profile);
int nautyextra_median_size (const graph *g, int m, int n, const set *profile);

int nautyextra_dilation (const graph *g, int m, int n);
int nautyextra_dilation_sg (const sparsegraph *sg);
int nautyextra_dilation_vpar (const int *vpar, int n);
int nautyextra_bandwidth (const graph *g, int m, int n);
int nautyextra_bandwidth_vpar (const int *vpar_orig, int n);

unsigned long long nautyextra_factorial (int n);
void nautyextra_numtoperm (NAUTYEXTRA_TYPE_ULL num, int n, int *perm);
void nautyextra_numtoperm1 (NAUTYEXTRA_TYPE_ULL num, int n, int *perm);
int nautyextra_is_idperm (const int *perm, int n);

int nautyextra_is_asymmetric(const graph *g, int m, int n);


#ifdef __cplusplus
}
#endif

#endif
